# -*- coding: utf-8 -*-
# Standard library imports
import os
import ssl
from json import dumps as json_dump
from json import loads as json_load

try:
    from urllib2 import urlopen, HTTPError, URLError
    from httplib import HTTPException
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
    from http.client import HTTPException

try:
    import certifi
except ImportError:
    from .vendor import certifi


def get(url):
    """Make a get request."""

    context = ssl.create_default_context(cafile=ca_certs())
    response = urlopen(url, context=context)
    return response


def json(response):
    """Get dict from json response."""

    return json_load(response.read().decode())


def ca_certs():
    """Returns path to vendored certifi/cacert.pem."""

    package = os.path.dirname(__file__)
    return os.path.join(package, "vendor", "certifi", "cacert.pem")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import os
import warnings
from bisect import bisect
from collections import OrderedDict

# Local imports
from . import compat, hooks, paths, repos
from .module import Module, ModuleSpec, module_header, sort_modules
from .resolver import Activator, Copier, Localizer, ResolveError, Resolver
from .vendor import appdirs, yaml

__all__ = [
    "activate",
    "deactivate",
    "clone",
    "create",
    "localize",
    "publish",
    "resolve",
    "set_home_path",
    "get_home_path",
    "get_home_modules_path",
    "get_cache_path",
    "get_user_path",
    "get_user_modules_path",
    "get_modules",
    "get_module_paths",
    "add_module_path",
    "get_active_modules",
    "add_active_module",
    "remove_active_module",
    "get_repos",
    "get_repo",
    "add_repo",
    "get_config_path",
    "read_config",
    "write_config",
    "update_repo",
    "remove_repo",
]
_registry = {
    "repos": OrderedDict(),
}
_active_modules = []
missing = object()


def resolve(requirements):
    """Resolve a list of module requirements."""

    resolver = Resolver(get_repos())
    return resolver.resolve(requirements)


def localize(requirements, to_repo="home", overwrite=False):
    """Localize a list of requirements."""

    to_repo = get_repo(to_repo)

    # Resolve modules
    resolver = Resolver(get_repos())
    module_specs = resolver.resolve(requirements)

    # Localize modules from remote repos
    localizer = Localizer(to_repo)
    modules = localizer.localize(module_specs, overwrite)

    return modules


def activate(requirements):
    """Resolve and active a list of module requirements.

    Usage:
        >>> cpenv.activate('moduleA', 'moduleB')

    Arguments:
        requirements (List[str]): List of module requirements

    Returns:
        list of Module objects that have been activated
    """

    # Resolve modules
    resolver = Resolver(get_repos())
    module_specs = resolver.resolve(requirements)

    # Activate modules
    activator = Activator()
    modules = activator.activate(module_specs)

    return modules


def activate_environment(environment):
    """Activate an environment by name.

    Usage:
        >>> cpenv.activate_environment('MyEnvironment')

    Arguments:
        environment (str): Name of Environment

    Returns:
        list of Module objects that have been activated
    """

    for repo in get_repos():
        for env in repo.list_environments():
            if env.name == environment:
                return activate(env.requires)
    else:
        raise ResolveError("Failed to resolve Environment: %s" % environment)


def deactivate():
    """Deactivates an environment by restoring all env vars to a clean state
    stored prior to activating environments
    """
    # TODO:
    # Probably need to store a clean environment prior to activate.
    # In practice it's uncommon to activate then deactivate in the same
    # python session.
    pass


def create(where, name, version, **kwargs):
    """Create a new module.

    Arguments:
        where (str): Path to new module
        name (str): Name of module
        version (str): Version of module
        description (str): Optional description of module
        author (str): Optional author of module
        email (str): Optional email address of author
        requires (list): Optional modules that this module depends on
        environment (dict): Optional environment variables

    Returns:
        Module object
    """

    # Setup configuration defaults
    config = OrderedDict(
        [
            ("name", name),
            ("version", version),
            ("description", kwargs.get("description", "")),
            ("author", kwargs.get("author", "")),
            ("email", kwargs.get("email", "")),
            ("requires", kwargs.get("requires", [])),
            ("environment", kwargs.get("environment", {})),
        ]
    )

    # Check if module already exists
    where = paths.normalize(where)
    if os.path.isdir(where):
        raise OSError('Module already exists at "%s"' % where)

    # Create a Module object - does not yet exist on disk
    module = Module(where, name, version)

    # Run global precreate hook
    # Allows users to inject data into a config prior to creating a new module
    hooks.run_global_hook("pre_create", module, config)

    # Create module folder structure
    paths.ensure_path_exists(where)
    paths.ensure_path_exists(where + "/hooks")

    data = module_header + yaml.dump(
        dict(config),
        default_flow_style=False,
        sort_keys=False,
    )
    with open(paths.normalize(where, "module.yml"), "w") as f:
        f.write(data)

    # Run global postcreate hook
    # Allows users to perform some action after a module is created
    hooks.run_global_hook("post_create", module)

    return module


def remove(module, from_repo=None):
    """Remove a module."""

    if isinstance(module, Module):
        return module.remove()

    if isinstance(module, ModuleSpec):
        return module.repo.remove(module)

    if from_repo is None:
        raise ValueError("from_repo is required when removing module by name.")

    if isinstance(from_repo, compat.string_types):
        from_repo = get_repo(from_repo)

    module_spec = from_repo.find(module)[0]
    return from_repo.remove(module_spec)


def clone(module, from_repo=None, where=None, overwrite=False):
    """Clone a module for local development.

    A typical development workflow using clone and publish:
        1. clone a module
        2. make changes
        3. test changes
        4. increment version in module.yml
        5. publish a new version of your module
    """

    if not isinstance(module, (Module, ModuleSpec)):
        if from_repo is None:
            resolver = Resolver(get_repos())
            module_spec = resolver.resolve([module])[0]
        else:
            from_repo = get_repo(from_repo)
            module_spec = from_repo.find(module)[0]

    module = module_spec.repo.download(
        module_spec,
        where=paths.normalize(where or ".", module_spec.real_name),
        overwrite=overwrite,
    )

    return module


def publish(module, to_repo="home", overwrite=False):
    """Publish a module to the specified repository."""

    to_repo = get_repo(to_repo)

    if isinstance(module, compat.string_types):
        resolver = Resolver(get_repos())
        module = resolver.resolve([module])[0]

    if isinstance(module, ModuleSpec):
        if not isinstance(module.repo, repos.LocalRepo):
            raise ValueError("Can only from modules in local repos.")
        else:
            module = Module(module.path)

    published = to_repo.upload(module, overwrite)
    return published


def copy(module, from_repo, to_repo, overwrite=False):
    """Copy a module from one repo to another."""

    from_repo = get_repo(from_repo)
    to_repo = get_repo(to_repo)

    # Resolve module
    resolver = Resolver([from_repo])
    module_spec = resolver.resolve([module])[0]

    copier = Copier(to_repo)
    copied = copier.copy([module_spec], overwrite)
    return copied


def get_active_modules():
    """Returns a list of active :class:`Module` s"""

    return _active_modules


def add_active_module(module):
    """Add a module to CPENV_ACTIVE_MODULES environment variable.

    Arguments:
        module (Module): Module to add to CPENV_ACTIVE_MODULES
    """

    if module not in _active_modules:
        _active_modules.append(module)

    module_names = os.pathsep.join([m.real_name for m in _active_modules])
    os.environ["CPENV_ACTIVE_MODULES"] = str(module_names)


def remove_active_module(module):
    """Remove a module from CPENV_ACTIVE_MODULES environment variable.

    Arguments:
        module (Module): Module to remove from CPENV_ACTIVE_MODULES
    """

    if module in _active_modules:
        _active_modules.remove(module)

    module_names = os.pathsep.join([m.real_name for m in _active_modules])
    os.environ["CPENV_ACTIVE_MODULES"] = str(module_names)


def set_home_path(path):
    """Convenient function used to set the CPENV_HOME environment variable."""

    # Set new home path
    home = paths.normalize(path)
    os.environ["CPENV_HOME"] = home
    _init_home_path(home)

    # Add new LocalRepo
    update_repo(repos.LocalRepo("home", get_home_modules_path()))

    return home


def _init_home_path(home):
    folders = [
        home,
        paths.normalize(home, "modules"),
        paths.normalize(home, "environments"),
        paths.normalize(home, "cache"),
    ]
    for folder in folders:
        paths.ensure_path_exists(folder)


def get_home_path():
    """Returns the cpenv home directory.

    Default home paths:
        win - C:/ProgramData/cpenv
        mac - /Library/Application Support/cpenv OR /Library/Caches/cpenv
        linux - /usr/local/share/cpenv OR ~/.local/share/cpenv
    """

    home = os.getenv("CPENV_HOME")
    if home and paths.is_writable(home):
        if paths.is_writable(home):
            return home
        else:
            raise RuntimeError("Could not access CPENV_HOME: %s", home)

    home_default = appdirs.site_data_dir("cpenv", appauthor=False)
    if paths.is_writable(home_default):
        return home_default
    else:
        if compat.platform == "mac":
            # Fallback to /Library/Caches as it has open permissions
            # and it's still shared across all users. This is not ideal
            # however, people don't clear their cache often.
            fallback = "/Library/Caches/cpenv"
        elif compat.platform == "linux":
            # Fallback to user directory on linux.
            fallback = appdirs.user_data_dir("cpenv", appauthor=False)
        else:
            raise RuntimeError("Could not access default home: %s", home)
        message = (
            'Could not access default cpenv home directory "{default}".'
            'Falling back to "{home}". If you want to use a shared home dir '
            "for all users on your machine run the following commands:\n\n"
            "    sudo mkdir {default}\n"
            "    sudo chmod -R a+rwx {default}\n"
        ).format(default=home_default, home=fallback)
        warnings.warn(message)
        return fallback


def get_home_modules_path():
    """Return the modules directory within the cpenv home directory."""

    return paths.normalize(get_home_path(), "modules")


def get_cache_path(*parts):
    """Return the cpenv cache directory within the cpenv home directory.

    Arguments:
        *parts (str) - List of path parts to join with cache path
    """

    return paths.normalize(get_home_path(), "cache", *parts)


def _init_user_path(user):
    """Initialize user path."""

    folders = [
        user,
        paths.normalize(user, "modules"),
    ]
    for folder in folders:
        paths.ensure_path_exists(folder)


def get_user_path():
    """Returns the cpenv user directory.

    Default user paths:
        win - C:/Users/<username>/AppData/Roaming/cpenv
        mac - ~/Library/Application Support/cpenv
        linux - ~/.local/share/cpenv
    """

    user_default = appdirs.user_data_dir("cpenv", appauthor=False)
    user = paths.normalize(user_default)

    return user


def get_user_modules_path():
    """Returns the modules directory within the cpenv user directory.

    Default user paths:
        win - C:/Users/<username>/AppData/Roaming/cpenv/modules
        mac - ~/Library/Application Support/cpenv/modules
        linux - ~/.local/share/cpenv/modules
    """

    return paths.normalize(get_user_path(), "modules")


def get_module_paths():
    """Returns a list of paths used to lookup local modules.

    The list of lookup paths contains:
        1. use modules path
        2. home modules path
        3. paths in CPENV_MODULES environment variable
    """

    module_paths = [paths.normalize(os.getcwd()), get_user_modules_path()]

    cpenv_home_modules = get_home_modules_path()
    if cpenv_home_modules not in module_paths:
        module_paths.append(cpenv_home_modules)

    cpenv_modules = os.environ.get("CPENV_MODULES", "").split(os.pathsep)
    for module_path in cpenv_modules:
        if module_path:
            module_paths.append(module_path)

    return module_paths


def add_module_path(path):
    """Add an additional lookup path for local modules."""

    path = paths.normalize(path)
    module_paths = []

    # Get existing module lookup paths
    cpenv_modules = os.environ.get("CPENV_MODULES", "").split(os.pathsep)
    for module_path in cpenv_modules:
        if module_path:
            module_paths.append(module_path)

    # Add new module lookup path
    if path not in module_paths:
        module_paths.append(path)
        add_repo(repos.LocalRepo(path, path))

    # Persist in CPENV_MODULES
    os.environ["CPENV_MODULES"] = os.pathsep.join(module_paths)

    return module_paths


def get_modules(*requirements):
    """Returns a list of available modules."""

    if requirements:
        resolver = Resolver(get_repos())
        return sort_modules(resolver.resolve(requirements))

    modules = []

    for repo in get_repos():
        modules.extend(repo.list())

    return sort_modules(list(modules))


def update_repo(repo):
    """Update a registered repo."""

    _registry["repos"].update({repo.name: repo})


def add_repo(repo, priority=None):
    """Register a Repo.

    Arguments:
        priority (int): Override the Repos priority when adding.
    """

    if priority is not None:
        repo.priority = priority

    if repo.name not in _registry["repos"]:
        repos = list(_registry["repos"].values())
        insert_idx = bisect([r.priority for r in repos], repo.priority)
        if insert_idx == len(repos):
            _registry["repos"][repo.name] = repo
        else:
            repos.insert(insert_idx, repo)
            _registry["repos"] = OrderedDict([(repo.name, repo) for repo in repos])


def remove_repo(repo):
    """Unregister a Repo."""

    _registry["repos"].pop(repo.name, None)


def get_repo(name, **query):
    """Get a repo by specifying an attribute to lookup"""

    if isinstance(name, repos.Repo):
        return name

    query["name"] = name

    for repo in get_repos():
        if all([getattr(repo, k, False) == v for k, v in query.items()]):
            return repo


def get_repos():
    """Get a list of all registered Repos."""

    return list(_registry["repos"].values())


def get_config_path():
    return paths.normalize(get_home_path(), "config.yml")


def read_config(key=None, default=missing):
    """Read the whole config or a specific key from disk.

    Examples:
        # Read whole config
        config = read_config()

        # Read one key
        repos = read_config('repos', {})
    """

    config_path = get_config_path()
    if not os.path.isfile(config_path):
        return {}

    with open(config_path, "r") as f:
        config = yaml.safe_load(f.read()) or {}

    if not key:
        return config

    config_key = config.get(key, missing)
    if config_key is missing:
        if default is missing:
            raise KeyError("Config has no key: " + key)
        else:
            return default

    return config_key


def write_config(*args):
    """Write the whole config or a specific key to disk.

    Examples:
        # Write whole config
        write_config({'repos': {}})

        # Write one config key
        write_config('repos', {})
    """
    if len(args) == 1:
        config = args[0]
    elif len(args) == 2:
        config = read_config()
        config[args[0]] = args[1]
    else:
        raise ValueError("Expected 1 or 2 arguments got %s" % len(args))

    config_path = get_config_path()
    with open(config_path, "w") as f:
        f.write(yaml.dump(config))


def _init():
    """Responsible for initially configuraing cpenv."""

    _init_home_path(get_home_path())
    _init_user_path(get_user_path())

    # Register builtin repos
    cwd = repos.LocalRepo("cwd", paths.normalize(os.getcwd()))
    user = repos.LocalRepo("user", get_user_modules_path())
    home = repos.LocalRepo("home", get_home_modules_path())
    if cwd.path == home.path == user.path:
        builtin_repos = [home]
    elif cwd.path == home.path:
        builtin_repos = [home, user]
    elif cwd.path == user.path:
        builtin_repos = [user, home]
    elif user.path == home.path:
        builtin_repos = [cwd, home]
    else:
        builtin_repos = [cwd, user, home]
    for repo in builtin_repos:
        add_repo(repo)

    # Register additional repos from CPENV_MODULE_PATHS
    builtin_module_paths = [repo.path for repo in get_repos()]
    for path in get_module_paths():
        if path in builtin_module_paths:
            continue
        add_repo(repos.LocalRepo(path, path))

    # Register repos from config
    configured_repos = read_config("repos", {})
    for name, config in configured_repos.items():
        repo_type = config.pop("type")
        repo_cls = repos.registry[repo_type]
        try:
            add_repo(repo_cls(**config))
        except Exception as e:
            warnings.warn(
                "Failed to create %s repo named %s\nError: %s"
                % (
                    repo_type,
                    config["name"],
                    str(e),
                )
            )

    # Set _active_modules from CPENV_ACTIVE_MODULES
    unresolved = []
    resolver = Resolver(get_repos())
    active_modules = os.getenv("CPENV_ACTIVE_MODULES", "").split(os.pathsep)
    for module in active_modules:
        if module:
            try:
                resolved = resolver.resolve([module])[0]
                _active_modules.append(resolved)
            except ResolveError:
                unresolved.append(module)

    if unresolved:
        warnings.warn("Unable to resolve %s from $CPENV_ACTIVE_MODULES:" % unresolved)



# -*- coding: utf-8 -*-
"""isort:skip_file"""
__title__ = "cpenv"
__version__ = "0.5.32"
__author__ = "Dan Bradham"
__email__ = "danielbradham@gmail.com"
__url__ = "http://github.com/cpenv/cpenv"
__license__ = "MIT"
__description__ = "Cross-platform Python environment management."


# Local imports
from .api import *
from .module import *
from .repos import *
from .reporter import *
from .resolver import *
from .versions import *
from . import vendor
from . import paths
from . import compat
from . import mappings


# Initialize cpenv
from .api import _init

_init()



from cpenv import api
from cpenv.cli import core
from cpenv.module import is_partial_match, sort_modules


class List(core.CLI):
    """List active and available Modules."""

    def setup_parser(self, parser):
        parser.add_argument(
            "requirement",
            help="Space separated list of modules.",
            nargs="?",
            default=None,
        )
        parser.add_argument(
            "--repo",
            help="List modules only in this repo.",
            default=None,
        )
        parser.add_argument(
            "--verbose",
            "-v",
            help="Print more module info.",
            action="store_true",
        )

    def run(self, args):

        found_modules = False

        core.echo()
        active_modules = api.get_active_modules()
        if args.requirement:
            active_modules = [
                m for m in active_modules if is_partial_match(args.requirement, m)
            ]

        if active_modules:
            found_modules = True
            core.echo(
                core.format_columns(
                    "[*] Active",
                    [m.real_name for m in sort_modules(active_modules)],
                )
            )
            core.echo()

        repos = api.get_repos()
        if args.repo:
            repos = [api.get_repo(args.repo)]

        for repo in repos:
            if args.requirement:
                repo_modules = repo.find(args.requirement)
            else:
                repo_modules = repo.list()

            module_names = []
            for module in sort_modules(repo_modules):
                if module in active_modules:
                    module_names.append("* " + module.real_name)
                else:
                    module_names.append("  " + module.real_name)

            if module_names:
                found_modules = True
                if repo.name != repo.path:
                    header = repo.name + " - " + repo.path
                else:
                    header = repo.name
                core.echo(
                    core.format_columns(
                        "[ ] " + header,
                        module_names,
                        indent="  ",
                    )
                )
                core.echo()

        if not found_modules:
            core.echo("No modules are available.")



import cpenv
from cpenv import api, paths
from cpenv.cli import core


class Version(core.CLI):
    """Show version information."""

    def run(self, args):

        core.echo()
        core.echo(
            core.format_section(
                "Version Info",
                [
                    ("name", cpenv.__name__),
                    ("version", cpenv.__version__),
                    ("url", cpenv.__url__),
                    ("package", paths.parent(cpenv.__file__)),
                    ("path", api.get_module_paths()),
                ],
            )
        )
        core.echo()

        # List package versions
        dependencies = []
        try:
            import Qt

            dependencies.extend(
                [
                    ("Qt.py", Qt.__version__),
                    ("Qt Binding", Qt.__binding__ + "-" + Qt.__binding_version__),
                ]
            )
        except ImportError:
            pass

        if not dependencies:
            return

        core.echo(core.format_section("Dependencies", dependencies), end="\n\n")



import sys

from cpenv import api, reporter
from cpenv.cli import core
from cpenv.resolver import ResolveError


class Info(core.CLI):
    """Show Module info"""

    usage = "cpenv info [-h] [<modules>...]"

    def setup_parser(self, parser):
        parser.add_argument(
            "modules",
            help="Space separated list of modules.",
            nargs="*",
        )
        parser.add_argument(
            "--key",
            help=(
                'Specify the name of a key like "path" or "version". '
                "When provided this command will print the value of the key."
            ),
            default=None,
        )
        parser.add_argument(
            "--repo",
            help="Get repo info instead of module info.",
            action="store_true",
        )
        parser.add_argument(
            "--home",
            help="Get the path to the cpenv home directory.",
            action="store_true",
        )
        parser.add_argument(
            "--cache",
            help="Get the path to the cpenv cache directory.",
            action="store_true",
        )

    def _get_key(self, obj, key):
        try:
            return eval("obj." + key)
        except Exception as e:
            core.echo("Error: %s" % e)
            sys.exit(1)

    def run(self, args):
        if args.home:
            core.echo(api.get_home_path())
            return

        if args.cache:
            core.echo(api.get_cache_path())
            return

        # Key mode - print a key's value of a module or repo.
        if args.key:
            reporter.set_reporter(reporter.Reporter())
            if args.repo:
                objects = []
                for repo_name in args.modules:
                    repo = api.get_repo(repo_name)
                    if repo is None:
                        core.echo("Could not find repo %s" % repo_name)
                        sys.exit(1)
                    objects.append(repo)
            else:
                try:
                    objects = api.resolve(args.modules)
                except ResolveError:
                    sys.exit(1)

            value = self._get_key(objects[0], args.key)
            core.echo(value)
            return

        # Normal mode - resolve repos or modules and show info.
        core.echo()
        if args.repo:
            repos = []
            for repo_name in args.modules:
                repo = api.get_repo(repo_name)
                if repo is None:
                    core.echo("Could not find repo %s" % repo_name)
                    sys.exit(1)
                repos.append(repo)

            for repo in repos:
                core.echo(
                    core.format_section(
                        repo.name,
                        [
                            ("name", repo.name),
                            ("path", repo.path),
                            ("type", str(type(repo))),
                            ("type_name", repo.type_name),
                        ],
                    )
                )
                core.echo()
            return
        else:
            try:
                module_specs = api.resolve(args.modules)
            except ResolveError:
                sys.exit(1)

            core.echo()
            for spec in module_specs:
                core.echo(
                    core.format_section(
                        spec.qual_name,
                        [(k, str(v)) for k, v in sorted(spec._asdict().items())],
                    )
                )
                core.echo()
            return



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import fnmatch
import os
import subprocess

# Local imports
from . import compat


def binpath(*paths):
    """Like os.path.join but acts relative to this packages bin path."""

    package_root = os.path.dirname(__file__)
    return os.path.normpath(os.path.join(package_root, "bin", *paths))


def run(*args, **kwargs):
    """Returns True if successful, False if failure"""

    kwargs.setdefault("env", os.environ)
    kwargs.setdefault("shell", True)

    try:
        subprocess.check_call(" ".join(args), **kwargs)
        return True
    except subprocess.CalledProcessError:
        print("Error running: {}".format(args))
        return False


def get_shell():
    """Get an approriate shell to use for a subshell.

    1. Use value of CPENV_SUBSHELL env var
    2. Attempt to walk up process tree using psutil
    3. Use a default shell (May not match with parent shell)

    Default Shells:
        win: powershell.exe
        mac: bash
        linux: bash
    """

    # Get preferred shell from
    preferred_shell = os.getenv("CPENV_SUBSHELL")
    if preferred_shell:
        return preferred_shell

    # Attempt to find parent shell using psutil
    try:
        import psutil

        supported_shells = [
            "powershell.exe",
            "cmd.exe",
            "powershell",
            "cmd",
            "bash",
            "sh",
            "fish",
            "xonsh",
            "zsh",
            "csh",
        ]

        process_parents = psutil.Process().parents()
        for proc in process_parents[::-1]:
            exe = proc.exe()
            exe_name = os.path.basename(exe)
            if exe_name in supported_shells:
                return exe
    except ImportError:
        pass

    # Fallback to a default shell
    return {
        "win": "powershell.exe",
        "linux": "bash",
        "mac": "bash",
    }[compat.platform]


def get_prompt(shell, prefix, colored=True):
    """Generate a prompt with a given prefix

    linux/osx: [prefix] user@host cwd $
          win: [prefix] cwd:
    """

    if shell.endswith("cmd.exe"):
        return "{} $P$G".format(prefix)

    if shell.endswith("powershell.exe"):
        return "'{} PS ' + $(get-location) + '> '".format(prefix)

    if shell.endswith("bash"):
        if colored:
            return (
                "{} "  # White prefix
                "\\[\\033[01;32m\\]\\u@\\h\\[\\033[00m\\] "  # Green user@host
                "\\[\\033[01;34m\\]\\w $ \\[\\033[00m\\]"  # Blue cwd $
            ).format(prefix)
        return "{} \\u@\\h \\w $ ".format(prefix)


def get_subshell_command(prefix):
    """Return a command to launch a subshell"""

    shell = get_shell()
    try:
        disable_prompt = int(os.getenv("CPENV_DISABLE_PROMPT", 0))
    except:
        disable_prompt = 0

    prompt = get_prompt(shell, prefix)

    if shell.endswith("cmd.exe"):
        if not disable_prompt:
            os.environ["PROMPT"] = prompt
        return [shell, "/K"]

    if shell.endswith("powershell.exe"):
        args = [shell, "-NoExit", "-NoLogo"]
        if not disable_prompt:
            args += ["-Command", "function Prompt {%s}" % prompt]
        return args

    if shell.endswith("bash"):
        if not disable_prompt:
            os.environ["PROMPT"] = prompt
        return [shell, binpath("subshell.sh")]

    return [shell]


def launch(prefix="[*]"):
    """Launch a subshell"""

    shell_cmd = get_subshell_command(prefix)
    subprocess.call(shell_cmd, env=dict(os.environ))



import argparse
import os
import re

from cpenv import api, repos, shell
from cpenv.cli import core


class Repo(core.CLI):
    """
    Manage, Configure and list Repos.

    Repos are sources of Modules. A Repo can be local or remote. Modules
    stored in remote Repos will be localized to a local repo prior to
    activation or explicitly by using "cpenv localize".
    """

    def commands(self):
        return [
            ListRepo(self),
            AddRepo(self),
            RemoveRepo(self),
            EditRepos(self),
        ]


class ListRepo(core.CLI):
    """List configured repos"""

    name = "list"

    def run(self, args):

        core.echo()
        core.echo("Repos in order of resolution:")
        for i, repo in enumerate(api.get_repos()):
            if repo.path == repo.name:
                core.echo(
                    "  [{}] {}  {}".format(
                        i,
                        type(repo).__name__,
                        repo.path,
                    )
                )
            else:
                core.echo(
                    "  [{}] {}  {}  {}".format(
                        i,
                        type(repo).__name__,
                        repo.name,
                        repo.path,
                    )
                )


class AddRepo(core.CLI):
    """Add a new repo.

    LocalRepo:
      cpenv repo add --type=local custom --path=~/custom_repo

    ShotgunRepo:
        cpenv repo add --type=shotgun my_shotgun --base_url=https://my.shotgunstudio.com --script_name=cpenv --api_key=secret

    The order of the arguments is important. First you have the --type and --priority
    options, then name, and finally repo type specific arguments.
    """

    name = "add"

    def setup_parser(self, parser):
        parser.add_argument(
            "name",
            help="Name of the repo",
        )
        parser.add_argument(
            "--type",
            help="Type of repo",
            choices=["local", "shotgun"],
            default="local",
        )
        parser.add_argument(
            "--priority",
            help="Priority of repo - the lower the priority the earlier in the list of repos it will appear. For example a priority of 0 would force the repo to be used first for module resolution. (Defaults: local - 10, shotgun - 20.)",
            default=None,
            type=int,
        )
        parser.add_argument(
            "type_args",
            help="Type specific arguments.",
            type=str,
            nargs=argparse.REMAINDER,
        )

    def parse_type_args(self, type_args):
        pattern = r"-{1,2}(?P<param>[a-zA-Z0-9_]+)=*\s*" r"[\'\"]?(?P<value>.+)[\'\"]?"
        kwargs = {}
        for arg in type_args:
            match = re.search(pattern, arg)
            param = match.group("param")
            raw_value = '"%s"' % match.group("value").strip()
            value = core.safe_eval(raw_value)
            kwargs[param] = value
        return kwargs

    def run(self, args):
        # Parse type_args or args that are specific to a given Repo type
        repo_kwargs = self.parse_type_args(args.type_args)
        repo_type = repo_kwargs.pop("type", args.type)

        core.echo()
        if repo_type not in repos.registry:
            core.echo("Error: %s is not a registered repo type." % args.type)
            core.echo("Choose from: " + ", ".join(repos.registry.keys()))
            core.exit(1)

        if api.get_repo(args.name):
            core.echo("Error: Repo named %s already exists." % args.name)
            core.exit(1)

        repo_cls = repos.registry[repo_type]
        repo_kwargs["name"] = args.name
        repo_kwargs["priority"] = args.priority
        core.echo("- Checking %s args..." % repo_cls.__name__, end="")
        try:
            repo_cls(**repo_kwargs)
        except Exception as e:
            core.echo("OOPS!")
            core.echo()
            core.echo("Error: Failed to initialize %s" % repo_cls.__name__)
            core.echo(str(e))
            core.exit(1)
        core.echo("OK!")
        core.echo()

        core.echo("- Adding repo to config...", end="")
        repo_kwargs["type"] = repo_type
        repo_config = api.read_config("repos", {})
        repo_config[args.name] = repo_kwargs
        api.write_config("repos", repo_config)
        core.echo("OK!")
        core.echo()


class RemoveRepo(core.CLI):
    """Remove a repo by name."""

    name = "remove"

    def setup_parser(self, parser):
        parser.add_argument(
            "name",
            help="Name of the repo",
        )

    def run(self, args):
        core.echo()
        if args.name in ["home", "user", "cwd"]:
            core.echo("Error: Can not remove %s repo." % args.name)

        repo_config = api.read_config("repos", {})
        if args.name not in repo_config:
            core.echo("Error: Repo named %s not found." % args.name)
            core.exit(1)

        core.echo("- Removing repo from config...", end="")
        repo_config.pop(args.name)
        api.write_config("repos", repo_config)
        core.echo("OK!")
        core.echo()


class EditRepos(core.CLI):
    """Open repos in text editor."""

    name = "edit"

    def run(self, args):

        config_path = api.get_config_path()
        editor = os.getenv("CPENV_EDITOR", os.getenv("EDITOR", "subl"))
        core.echo("Opening %s in %s." % (config_path, editor))
        shell.run(editor, config_path)



from cpenv import api
from cpenv.cli import core
from cpenv.resolver import Copier, Resolver


class Copy(core.CLI):
    """Copy Modules from one Repo to another."""

    def setup_parser(self, parser):
        parser.add_argument(
            "modules",
            help="Space separated list of modules.",
            nargs="+",
        )
        parser.add_argument(
            "--from_repo",
            help="Download from",
            default=None,
        )
        parser.add_argument(
            "--to_repo",
            help="Upload to",
            default=None,
        )
        parser.add_argument(
            "--overwrite",
            help="Overwrite the destination directory. (False)",
            action="store_true",
        )

    def run(self, args):

        core.echo()

        if args.from_repo:
            from_repo = api.get_repo(args.from_repo)
        else:
            from_repo = core.prompt_for_repo(
                api.get_repos(),
                "Download from",
                default_repo_name="home",
            )

        if args.to_repo:
            to_repo = api.get_repo(args.to_repo)
        else:
            repos = api.get_repos()
            repos.remove(from_repo)
            to_repo = core.prompt_for_repo(
                repos,
                "Upload to",
                default_repo_name="home",
            )

        resolver = Resolver([from_repo])
        module_specs = resolver.resolve(args.modules)
        core.echo()

        choice = core.prompt("Copy these modules to %s?[y/n] " % to_repo.name)
        if choice.lower() not in ["y", "yes", "yup"]:
            core.echo("Aborted.")
            core.exit(1)
        core.echo()

        copier = Copier(to_repo)
        copier.copy(module_specs, args.overwrite)



# -*- coding: utf-8 -*-


class Repo(object):
    """Base class for all Repos.

    A Repo is a source of modules. They can be local or remote so long as they
    provide this interface.
    """

    type_name = "repo"
    priority = 10

    def __init__(self, name, priority=None):
        self.name = name
        self.priority = self.priority if priority is None else priority

    def __repr__(self):
        args = []
        for attr in ["name", "path"]:
            if hasattr(self, attr):
                args.append("{}={!r}".format(attr, getattr(self, attr)))
        return "<{}>({})".format(type(self).__name__, ", ".join(args))

    def clear_cache(self):
        """Subclasses that implement caching should implement this method
        so that cpenv can clear the cache when necessary. Like after copying
        or localizing modules.
        """
        return NotImplemented

    def find(self, requirement):
        """Given a requirement, return a list of ModuleSpecs that match.

        Should be ordered from best to worst match.
        """

        return NotImplemented

    def list(self):
        """Return a list of ModuleSpecs in this Repo."""

        return NotImplemented

    def download(self, module_spec, where, overwrite=False):
        """Given a module_spec and output path, download it from this Repo.

        Return a newly downloaded Module.
        """

        return NotImplemented

    def upload(self, module, overwrite=False):
        """Given a module, upload it to this Repo.

        Return a ModuleSpec.
        """

        return NotImplemented

    def remove(self, module_spec):
        """Given a module_spec, remove it from this repo."""

        return NotImplemented

    def get_data(self, module_spec):
        """Given a module_spec, return a Module's config(module.yml) dict.

        If possible this should be implemented for all Repos, however, it
        is not used during module resolution. It is only used to provide
        a richer view of Modules in Repos by UI implementations.
        """

        return {}

    def get_size(self, module_spec):
        """Given a module_spec, return the size of a module in the repo.

        For remote repositories this may mean reading a field in a databse.
        For the LocalRepo it means getting the combined size of the module
        folder's contents.

        Repos should implement this method to full support progress bars
        while publishing and localizing modules.

        Return -1 when size if the size is unavailable.
        """
        return -1

    def get_thumbnail(self, module_spec):
        """Given a module_spec, return the path to a thumbnail.

        Remote repos should download a thumbnail to the cache/icons:
            cache_path = cpenv.get_cache_path('icons', 'cached-0.1.0.png')

        Return None when no thumbnail is available.
        """
        return

    def list_environments(self, filters=None):
        """Return a list of Environments in this repo.

        Filters should be a dict containing fields to match in the list of
        results for example:
            {
                'name': 'My*',
                'requires': ['my_module'],
            }

        The above filters should return a list of Environments with names
        beginning with My and requires containing my_module.
        """

        return []

    def save_environment(self, name, data, force=False):
        """Save an environment to this repo.

        This method needs to store an Environment within the Repo. This could
        mean saving to a database or to a file on disk in the case of
        local repositories.
        """

        return NotImplemented

    def remove_environment(self, name):
        """Removes an Environment from this repo."""

        return NotImplemented



# -*- coding: utf-8 -*-

# Standard library imports
import re
from collections import namedtuple
from functools import total_ordering

# Local imports
from . import compat

__all__ = [
    "ParseError",
    "Version",
    "parse_version",
    "default_version",
]


# Strange nuke version pattern
nuke_version_pattern = r"(?P<major>\d+)" r"\.(?P<minor>\d+)" r"(?:v)" r"(?P<patch>\d+)$"

# Version pattern with 4 digits
four_version_pattern = (
    r"(?:v)?"
    r"(?P<major>\d+)"
    r"\.(?P<minor>\d+)"
    r"\.(?P<revision>\d+)"
    r"\.(?P<build>\d+)"
    r"(?:-(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)

# Modified regex from semver.org - works with calver as well
semver_version_pattern = (
    r"(?:v)?"
    r"(?P<major>\d+)"
    r"\.(?P<minor>\d+)"
    r"\.(?P<patch>\d+)"
    r"(?:-(?P<prerelease>(?:\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
    r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)
simplever_pattern = r"(?:v)?(?P<version>(\d+\.?)+)$"
VersionBase = namedtuple(
    "Version", ["major", "minor", "patch", "prerelease", "buildmetadata", "string"]
)


@total_ordering
class Version(VersionBase):

    _defaults = {
        "major": 0,
        "minor": 0,
        "patch": 0,
        "prerelease": None,
        "buildmetadata": None,
    }

    def __str__(self):
        return self.string

    def __hash__(self):
        return super(Version, self).__hash__()

    def _comparable_value(self, value, other):
        """Return a value that will be comparable with other."""

        types_match = type(value) is type(other)
        if types_match or isinstance(value, compat.numeric_types):
            return value

        if other is None:
            return value

        if value is None and isinstance(other, compat.string_types):
            return "zzzzzzzz"

        if isinstance(other, compat.numeric_types):
            return -float("inf")

        return value

    def _comparable(self, other):
        """Generate a comparison key for a Version object.

        Coerces types of prerelease and buildmeta to ensure that the Version
        objects are comparable. This is required because Python 3 now raises
        a TypeError when attempting to compare str and int.
        """
        return (
            self.major,
            self.minor,
            self.patch,
            self._comparable_value(self.prerelease, other.prerelease),
            self._comparable_value(self.buildmetadata, other.buildmetadata),
        )

    def __lt__(self, other):
        if not isinstance(other, Version):
            raise ValueError("Can only compare two Version objects.")

        return self._comparable(other) < other._comparable(self)

    def __eq__(self, other):
        if not isinstance(other, Version):
            raise ValueError("Can only compare two Version objects.")

        return tuple(other) == tuple(self)


class ParseError(Exception):
    """Raised when a parse method fails."""


def parse_version(string):
    """Parse and return a Version from the provided string.

    Supports:
      - semver / calver
      - simple versions like: 10, v2, 1.0, 2.2.4

    Arguments:
        string (str): String to parse version from.

    Returns:
        Version object

    Raises:
        ParseError when a version can not be parsed.
    """
    # Parse weird nuke versioning
    match = re.search(nuke_version_pattern, string)
    if match:
        return Version(
            major=int(match.group("major")),
            minor=int(match.group("minor")),
            patch=int(match.group("patch")),
            prerelease=None,
            buildmetadata=None,
            string=match.group(0),
        )

    # Parse four_version - four digit version
    match = re.search(four_version_pattern, string)
    if match:
        return Version(
            major=int(match.group("major")),
            minor=int(match.group("minor")),
            patch=int(match.group("revision")),
            prerelease=int(match.group("build")),
            buildmetadata=match.group("buildmetadata"),
            string=match.group(0),
        )

    # Parse Semver / Calver
    match = re.search(semver_version_pattern, string)
    if match:
        return Version(
            major=int(match.group("major")),
            minor=int(match.group("minor")),
            patch=int(match.group("patch")),
            prerelease=match.group("prerelease"),
            buildmetadata=match.group("buildmetadata"),
            string=match.group(0),
        )

    # Parse Simple version
    match = re.search(simplever_pattern, string)
    if match:
        kwargs = dict(Version._defaults)
        kwargs["string"] = match.group(0)
        version_parts = match.group("version").split(".")
        for part, part_name in zip(version_parts, ["major", "minor", "patch"]):
            kwargs[part_name] = int(part)
        return Version(**kwargs)

    raise ParseError("Could not parse version from %s" % string)


def default_version():
    return Version(
        major=0,
        minor=1,
        patch=0,
        prerelease=None,
        buildmetadata=None,
        string="0.1.0",
    )



# -*- coding: utf-8 -*-

# Standard library imports
import os
import sys

py_ver = sys.version_info[0]
is_py2 = py_ver == 2
is_py3 = py_ver == 3

if is_py2:
    string_types = (str, basestring, unicode)
    numeric_types = (int, long, float)

if is_py3:
    string_types = (str, bytes)
    numeric_types = (int, float)


platform = sys.platform.rstrip("1234567890").lower()
if platform == "darwin":  # Use mac instead of darwin
    platform = "mac"
os.environ["CPENV_PLATFORM"] = platform


# Leave osx in supported platforms for backwards compatability
supported_platforms = ["win", "linux", "mac", "osx"]



# -*- coding: utf-8 -*-
"""
Tools for working with paths and files.
"""

# Standard library imports
import os
import shutil
import stat
import zipfile
from fnmatch import fnmatch


def normalize(*parts):
    """Join, expand, and normalize a filepath."""

    path = os.path.expanduser(os.path.expandvars(os.path.join(*parts)))
    return os.path.abspath(path).replace("\\", "/")


def parent(path):
    """Get a file paths parent directory"""
    return normalize(os.path.dirname(path))


def ensure_path_exists(path, *args):
    """Like os.makedirs but keeps quiet if path already exists"""

    if os.path.exists(path):
        return

    os.makedirs(path, *args)


def is_writable(path):
    """Check if a directory is writable."""

    if os.path.exists(path):
        return os.access(path, os.X_OK | os.W_OK)

    try:
        os.makedirs(path)
        touch(normalize(path, "tmpfile"))
    except OSError as e:
        return False
    else:
        rmtree(path)
        return True


def rmtree(path):
    """Safely remove directory and all of it's contents."""

    def onerror(func, path, _):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(path, onerror=onerror)


def walk_up(start_dir, depth=20):
    """Like os.walk but walks up a tree."""

    root = start_dir

    for i in range(depth):
        contents = os.listdir(root)
        subdirs, files = [], []
        for f in contents:
            if os.path.isdir(os.path.join(root, f)):
                subdirs.append(f)
            else:
                files.append(f)

        yield root, subdirs, files

        parent = os.path.dirname(root)
        if parent and not parent == root:
            root = parent
        else:
            break


def touch(filepath):
    """Touch the given filepath"""

    with open(filepath, "a"):
        os.utime(filepath, None)


def format_size(bytesize):
    """Human readable size."""

    value = bytesize
    for unit in ["b", "kb", "mb", "gb", "tb", "pb", "zi"]:
        if value < 1024.0:
            return "{:.0f} {}".format(value, unit)
        value /= 1024.0
    return "{:.0f} {}".format(value, unit)


def get_file_count(folder):
    """Get the number of files in a folder."""

    count = 0
    for _, _, files in exclusive_walk(folder):
        count += len(files)
    return count


def get_folder_size(folder):
    """Get the size of a folder in bytes."""

    size = 0
    for root, _, files in exclusive_walk(folder):
        for file in files:
            file_path = os.path.join(root, file)
            if not os.path.islink(file_path):
                size += os.path.getsize(file_path)
    return size


def exclude_names(names):
    """Returns True when a file matches one of the provided names."""

    def check_file_against_names(file):
        return os.path.basename(file) in names

    return check_file_against_names


def exclude_patterns(patterns):
    """Returns True when a file matches against one of the provided patterns"""

    def check_file_against_patterns(file):
        return any([fnmatch(file, p) for p in patterns])

    return check_file_against_patterns


def include_prebuilt_pyc(file):
    """Returns True when a pyc file has no accompanying py file.

    This is useful in cases where tool distributors prebuild pyc files as
    a crude form of anti-piracy. These need to be included in cpenv modules
    when zipped to be uploaded to a repository.
    """

    return (
        file.endswith(".pyc")
        and not os.path.isfile(file.replace(".pyc", ".py"))
        and "__pycache__" not in file
    )


def is_excluded(value, predicates):
    """Check if a value matches any of the exclude predicates."""

    return any([predicate(value) for predicate in predicates])


def is_included(value, predicates):
    """Returns True if value matches any of the include predicates."""

    return any([predicate(value) for predicate in predicates])


def exclusive_walk(folder, excludes=None, includes=None):
    """Like os.walk but excludes/includes files by using predicate functions.

    Excludes the following by default:
        names: __pycache__, .git, Thumbs.db, .venv, venv
        patterns: *.pyc, *.egg-info

    Includes the following by default:
        .pyc files that have no accompanying .py file.

    Arguments:
        folder (str): Root folder to recursively walk.
        excludes ([callable]): List of predicate functions used to exclude files.
        includes ([callable]): List of predicate functions to include files. Overrides excludes.

    Returns:
        Generator yielding (root, subdirs, files).
    """

    excludes = excludes or [
        exclude_names(["__pycache__", ".git", "thumbs.db", ".venv", "venv"]),
        exclude_patterns(["*.pyc", "*.egg-info"]),
    ]
    includes = includes or [include_prebuilt_pyc]

    for root, subdirs, files in os.walk(folder):
        if is_excluded(root, excludes) and not is_included(root, includes):
            subdirs[:] = []
            continue

        included_files = []
        for file in files:
            path = normalize(root, file)
            if is_excluded(path, excludes) and not is_included(path, includes):
                continue
            included_files.append(file)

        yield root, subdirs, included_files


def zip_folder(folder, where):
    """Zip the contents of a folder."""

    parent = os.path.dirname(where)
    if not os.path.isdir(parent):
        os.makedirs(parent)

    # TODO: Count files first so we can report progress of building zip.

    with zipfile.ZipFile(where, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for root, subdirs, files in exclusive_walk(folder):
            rel_root = os.path.relpath(root, folder)
            for file in files:
                zip_file.write(
                    os.path.join(root, file), arcname=os.path.join(rel_root, file)
                )



from cpenv import api
from cpenv.cli import core
from cpenv.module import best_match


class Remove(core.CLI):
    """Permanently remove a Module from a Repo."""

    def setup_parser(self, parser):
        parser.add_argument(
            "modules",
            help="Space separated list of modules.",
            nargs="+",
        )
        parser.add_argument(
            "--from_repo",
            help="Specific repo to remove from.",
            default=None,
        )

    def run(self, args):

        core.echo()

        if args.from_repo:
            from_repo = api.get_repo(name=args.from_repo)
        else:
            from_repo = core.prompt_for_repo(
                api.get_repos(),
                "Choose a repo to remove module from",
                default_repo_name="home",
            )

        core.echo("- Finding modules in %s..." % from_repo.name)
        core.echo()
        modules_to_remove = []
        for requirement in args.modules:
            module = best_match(requirement, from_repo.find(requirement))
            if not module:
                core.echo("Error: %s not found..." % requirement)
                core.exit(1)
            core.echo("  %s - %s" % (module.real_name, module.path))
            modules_to_remove.append(module)

        core.echo()
        choice = core.prompt("Delete these modules?[y/n] ")
        if choice.lower() not in ["y", "yes", "yup"]:
            core.echo("Aborted.")
            core.exit(1)

        core.echo()
        core.echo("- Removing modules...")
        core.echo()
        for module in modules_to_remove:
            core.echo("  " + module.real_name)
            api.remove(module, from_repo)

        core.echo()
        core.echo("Successfully removed modules.")
        core.echo()



import os

from cpenv import api, paths
from cpenv.cli import core
from cpenv.module import parse_module_path


class Create(core.CLI):
    """Create a new Module."""

    def setup_parser(self, parser):
        parser.add_argument(
            "where",
            help="Path to new module",
        )

    def run(self, args):
        where = paths.normalize(args.where)
        if os.path.isdir(where):
            core.echo()
            core.echo("Error: Can not create module in existing directory.")
            core.exit(1)

        default_name, default_version = parse_module_path(where)

        core.echo()
        core.echo("This command will guide you through creating a new module.")
        core.echo()
        name = core.prompt("  Module Name [%s]: " % default_name)
        version = core.prompt("  Version [%s]: " % default_version.string)
        description = core.prompt("  Description []: ")
        author = core.prompt("  Author []: ")
        email = core.prompt("  Email []: ")

        core.echo()
        core.echo("- Creating your new Module...", end="")
        module = api.create(
            where=where,
            name=name or default_name,
            version=version or default_version.string,
            description=description,
            author=author,
            email=email,
        )
        core.echo("OK!")

        core.echo()
        core.echo("  " + module.path)

        core.echo()
        core.echo("Steps you might take before publishing...")
        core.echo()
        core.echo("  - Include binaries your module depends on")
        core.echo("  - Edit the module.yml file")
        core.echo("    - Add variables to the environment section")
        core.echo("    - Add other modules to the requires section")
        core.echo("  - Add python hooks like post_activate")
        core.echo()
